@@ -1,17 +1,14 @@
 require 'mkmf'
 
-$CPPFLAGS += " -Dss_family=__ss_family -Dss_len=__ss_len"
-
 case RUBY_PLATFORM
+when /bccwin32/
+  test_func = "WSACleanup"
+  have_library("ws2_32", "WSACleanup")
 when /mswin32|mingw/
   test_func = "WSACleanup"
   have_library("wsock32", "WSACleanup")
-  have_func("closesocket")
 when /cygwin/
-#  $LDFLAGS << " -L/usr/lib" if File.directory?("/usr/lib")
-#  $CFLAGS << " -I/usr/include"
   test_func = "socket"
-#  have_library("bind", "gethostbyaddr")
 when /beos/
   test_func = "socket"
   have_library("net", "socket")
@@ -25,9 +22,18 @@ else
   have_library("socket", "socket")
 end
 
+unless $mswin or $bccwin or $mingw
+  headers = %w<sys/types.h netdb.h string.h sys/socket.h netinet/in.h>
+end
+if /solaris/ =~ RUBY_PLATFORM and !try_compile("")
+  # bug of gcc 3.0 on Solaris 8 ?
+  headers << "sys/feature_tests.h"
+end
+
 $ipv6 = false
-if enable_config("ipv6", false)
-  if try_link(<<EOF)
+default_ipv6 = /cygwin/ !~ RUBY_PLATFORM
+if enable_config("ipv6", default_ipv6)
+  if checking_for("ipv6") {try_link(<<EOF)}
 #include <sys/types.h>
 #include <sys/socket.h>
 main()
@@ -45,19 +51,13 @@ $ipv6lib = nil
 $ipv6libdir = nil
 $ipv6trylibc = nil
 if $ipv6
-  if egrep_cpp("yes", <<EOF)
+  if macro_defined?("IPV6_INRIA_VERSION", <<EOF)
 #include <netinet/in.h>
-#ifdef IPV6_INRIA_VERSION
-yes
-#endif
 EOF
     $ipv6type = "inria"
     $CFLAGS="-DINET6 "+$CFLAGS
-  elsif egrep_cpp("yes", <<EOF)
+  elsif macro_defined?("__KAME__", <<EOF)
 #include <netinet/in.h>
-#ifdef __KAME__
-yes
-#endif
 EOF
     $ipv6type = "kame"
     $ipv6lib="inet6"
@@ -69,31 +69,22 @@ EOF
     $ipv6lib="inet6"
     $ipv6libdir="/usr/inet6/lib"
     $CFLAGS="-DINET6 -I/usr/inet6/include "+$CFLAGS
-  elsif egrep_cpp("yes", <<EOF)
+  elsif macro_defined?("_TOSHIBA_INET6", <<EOF)
 #include <sys/param.h>
-#ifdef _TOSHIBA_INET6
-yes
-#endif
 EOF
     $ipv6type = "toshiba"
     $ipv6lib="inet6"
     $ipv6libdir="/usr/local/v6/lib"
     $CFLAGS="-DINET6 "+$CFLAGS
-  elsif egrep_cpp("yes", <<EOF)
+  elsif macro_defined?("__V6D__", <<EOF)
 #include </usr/local/v6/include/sys/v6config.h>
-#ifdef __V6D__
-yes
-#endif
 EOF
     $ipv6type = "v6d"
     $ipv6lib="v6"
     $ipv6libdir="/usr/local/v6/lib"
     $CFLAGS="-DINET6 -I/usr/local/v6/include "+$CFLAGS
-  elsif egrep_cpp("yes", <<EOF)
+  elsif macro_defined?("_ZETA_MINAMI_INET6", <<EOF)
 #include <sys/param.h>
-#ifdef _ZETA_MINAMI_INET6
-yes
-#endif
 EOF
     $ipv6type = "zeta"
     $ipv6lib="inet6"
@@ -120,79 +111,47 @@ EOS
   end
 end
 
-  if try_link(<<EOF)
-#ifdef _WIN32
-# include <windows.h>
-# include <winsock.h>
-#else
-# include <sys/types.h>
-# include <netdb.h>
-# include <string.h>
-# include <sys/socket.h>
-# include <netinet/in.h>
-#endif
-int
-main()
-{
-   struct sockaddr_in sin;
-
-   sin.sin_len;
-   return 0;
-}
-EOF
-    $CFLAGS="-DHAVE_SIN_LEN "+$CFLAGS
+if have_struct_member("struct sockaddr_in", "sin_len", headers)
+  $defs[-1] = "-DHAVE_SIN_LEN"
 end
 
-  if try_link(<<EOF)
-#ifdef _WIN32
-# include <windows.h>
-# include <winsock.h>
-#else
-# include <sys/types.h>
-# include <netdb.h>
-# include <string.h>
-# include <sys/socket.h>
-#endif
-int
-main()
-{
-   struct sockaddr_storage ss;
-
-   ss.ss_family;
-   return 0;
-}
-EOF
-    $CFLAGS="-DHAVE_SOCKADDR_STORAGE "+$CFLAGS
+#   doug's fix, NOW add -Dss_family... only if required!
+[nil, " -Dss_family=__ss_family -Dss_len=__ss_len"].each do |flags|
+  if flags
+    cppflags = $CPPFLAGS
+    $CPPFLAGS += flags
+  end
+  if have_struct_member("struct sockaddr_storage", "ss_family", headers)
+    $defs[-1] = "-DHAVE_SOCKADDR_STORAGE"
+    break
+  elsif flags
+    $CPPFLAGS = cppflags
+  end
 end
 
-  if try_link(<<EOF)
-#include <sys/types.h>
-#include <netdb.h>
-#include <string.h>
-#include <sys/socket.h>
-#include <netinet/in.h>
-int
-main()
-{
-   struct sockaddr sa;
-
-   sa.sa_len;
-   return 0;
-}
-EOF
-    $CFLAGS="-DHAVE_SA_LEN "+$CFLAGS
+if have_struct_member("struct sockaddr", "sa_len", headers)
+  $defs[-1] = "-DHAVE_SA_LEN "
 end
 
-have_header("netinet/tcp.h") if not /cygwin/ === RUBY_PLATFORM # for cygwin 1.1.5
+have_header("netinet/tcp.h") if not /cygwin/ =~ RUBY_PLATFORM # for cygwin 1.1.5
 have_header("netinet/udp.h")
 
-$getaddr_info_ok = false
-if not enable_config("wide-getaddrinfo", false) and try_run(<<EOF)
-#include <sys/types.h>
-#include <netdb.h>
-#include <string.h>
-#include <sys/socket.h>
-#include <netinet/in.h>
+if have_func("sendmsg") | have_func("recvmsg")
+  have_struct_member('struct msghdr', 'msg_control', ['sys/types.h', 'sys/socket.h'])
+  have_struct_member('struct msghdr', 'msg_accrights', ['sys/types.h', 'sys/socket.h'])
+end
+
+getaddr_info_ok = enable_config("wide-getaddrinfo") do
+  checking_for("wide getaddrinfo") {try_run(<<EOF)}
+#{cpp_include(headers)}
+#include <stdlib.h>
+
+#ifndef EXIT_SUCCESS
+#define EXIT_SUCCESS 0
+#endif
+#ifndef EXIT_FAILURE
+#define EXIT_FAILURE 1
+#endif
 
 #ifndef AF_LOCAL
 #define AF_LOCAL AF_UNIX
@@ -216,8 +175,13 @@ main()
     }
     for (ai = aitop; ai; ai = ai->ai_next) {
       if (ai->ai_family == AF_LOCAL) continue;
-      if (ai->ai_addr == NULL ||
-          ai->ai_addrlen == 0 ||
+      if (ai->ai_addr == NULL)
+        goto bad;
+#if defined(_AIX)
+      ai->ai_addr->sa_len = ai->ai_addrlen;
+      ai->ai_addr->sa_family = ai->ai_family;
+#endif
+      if (ai->ai_addrlen == 0 ||
           getnameinfo(ai->ai_addr, ai->ai_addrlen,
                       straddr, sizeof(straddr), strport, sizeof(strport),
                       NI_NUMERICHOST|NI_NUMERICSERV) != 0) {
@@ -268,51 +232,46 @@ main()
 
   if (aitop)
     freeaddrinfo(aitop);
-  exit(0);
+  exit(EXIT_SUCCESS);
 
  bad:
   if (aitop)
     freeaddrinfo(aitop);
-  exit(1);
+  exit(EXIT_FAILURE);
 }
 EOF
-  $getaddr_info_ok = true
 end
-if $ipv6 and not $getaddr_info_ok
-      print <<EOS
+if $ipv6 and not getaddr_info_ok
+  abort <<EOS
 
 Fatal: --enable-ipv6 is specified, and your OS seems to support IPv6 feature.
 But your getaddrinfo() and getnameinfo() are appeared to be broken.  Sorry,
 you cannot compile IPv6 socket classes with broken these functions.
+You can try --enable-wide-getaddrinfo.
 EOS
-  exit
 end
-      
+
 case with_config("lookup-order-hack", "UNSPEC")
 when "INET"
-  $CFLAGS="-DLOOKUP_ORDER_HACK_INET "+$CFLAGS
+  $defs << "-DLOOKUP_ORDER_HACK_INET"
 when "INET6"
-  $CFLAGS="-DLOOKUP_ORDER_HACK_INET6 "+$CFLAGS
+  $defs << "-DLOOKUP_ORDER_HACK_INET6"
 when "UNSPEC"
   # nothing special
 else
-  print <<EOS
+  abort <<EOS
 
 Fatal: invalid value for --with-lookup-order-hack (expected INET, INET6 or UNSPEC)
 EOS
-  exit
 end
 
 $objs = ["socket.#{$OBJEXT}"]
-    
-if $getaddr_info_ok and have_func("getaddrinfo") and have_func("getnameinfo")
-  have_getaddrinfo = true
-end
 
-if have_getaddrinfo
-  $CFLAGS="-DHAVE_GETADDRINFO "+$CFLAGS
-else
-  $CFLAGS="-I. "+$CFLAGS
+unless getaddr_info_ok and have_func("getnameinfo", "netdb.h") and have_func("getaddrinfo", "netdb.h")
+  if have_struct_member("struct in6_addr", "s6_addr8", headers)
+    $defs[-1] = "-DHAVE_ADDR8"
+  end
+  $CPPFLAGS="-I. "+$CPPFLAGS
   $objs += ["getaddrinfo.#{$OBJEXT}"]
   $objs += ["getnameinfo.#{$OBJEXT}"]
   have_func("inet_ntop") or have_func("inet_ntoa")
@@ -323,34 +282,25 @@ else
   have_header("resolv.h")
 end
 
-if !try_link(<<EOF)
-#include <sys/types.h>
-#include <netdb.h>
-#include <string.h>
-#include <sys/socket.h>
-#include <netinet/in.h>
-int
-main()
-{
-   socklen_t len;
-   return 0;
-}
-EOF
-  $CFLAGS="-Dsocklen_t=int "+$CFLAGS
+unless have_type("socklen_t", headers)
+  $defs << "-Dsocklen_t=int"
 end
 
 have_header("sys/un.h")
+have_header("sys/uio.h")
 
 if have_func(test_func)
   have_func("hsterror")
+  have_func("getipnodebyname") or have_func("gethostbyname2")
+  have_func("socketpair")
   unless have_func("gethostname")
     have_func("uname")
   end
-  if ENV["SOCKS_SERVER"] or enable_config("socks", false)
+  if enable_config("socks", ENV["SOCKS_SERVER"])
     if have_library("socks5", "SOCKSinit")
-      $CFLAGS+=" -DSOCKS5 -DSOCKS"
+      $defs << "-DSOCKS5" << "-DSOCKS"
     elsif have_library("socks", "Rconnect")
-      $CFLAGS+=" -DSOCKS"
+      $defs << "-DSOCKS"
     end
   end
   create_makefile("socket")
